from __future__ import annotations
from typing import Optional, ClassVar, Generator
import ctypes

from .libpysdcxx import libpysdcxx
from .util import serialise


class BigramMultiset(Generator[str, None, None]):
    """
    Bigram multiset (implemented by std::multiset)
    """

    _str_fixed_len: ClassVar[int] = len("wbigram_multiset(size: XXXXXXXXXX, {})")

    def __init__(self, string: Optional[str] = None, _impl: Optional = None):
        """
        :param string: String from which bigrams multiset shall be created
        """
        self._impl = libpysdcxx.new_wbigram_multiset_str(ctypes.c_wchar_p(string)) \
            if string is not None else _impl or libpysdcxx.new_wbigram_multiset()

    def __deepcopy__(self, memo):
        """
        Make a copy on the native level
        :param memo: IDs of already copied objects (unused, we're non-recursive)
        """
        return BigramMultiset(_impl=libpysdcxx.new_wbigram_multiset_copy(self._impl))

    def __copy__(self):
        """
        We don't do shallow copies
        """
        return self.__deepcopy__(None)

    def __len__(self):
        return libpysdcxx.wbigram_multiset_size(self._impl)

    def __iter__(self):
        """
        :return: Generator of bigrams together with their counts as tuple[str, int]
        """
        itr = libpysdcxx.wbigram_multiset_cbegin(self._impl)
        end = libpysdcxx.wbigram_multiset_cend(self._impl)
        try:
            while libpysdcxx.wbigram_multiset_citer_ne(itr, end):
                ch1, ch2 = ctypes.c_wchar(), ctypes.c_wchar()
                libpysdcxx.wbigram_multiset_citer_deref(
                    itr, ctypes.byref(ch1), ctypes.byref(ch2))

                yield ch1.value + ch2.value

                libpysdcxx.wbigram_multiset_citer_inc(itr)

        finally:
            libpysdcxx.delete_wbigram_multiset_citer(end)
            libpysdcxx.delete_wbigram_multiset_citer(itr)

    def send(self):
        pass

    def throw(self):
        pass

    def __iadd__(self, other: BigramMultiset) -> BigramMultiset:
        """
        Update by `other` bigrams (in-place union)
        """
        assert isinstance(other, BigramMultiset)
        libpysdcxx.wbigram_multiset_iadd(self._impl, other._impl)
        return self

    def __add__(self, other: BigramMultiset) -> BigramMultiset:
        """
        :return: Union of `self` and `other` bigrams
        """
        assert isinstance(other, BigramMultiset)
        return BigramMultiset(
            _impl=libpysdcxx.wbigram_multiset_add(self._impl, other._impl))

    @staticmethod
    def intersect_size(bgrms1: BigramMultiset, bgrms2: BigramMultiset) -> int:
        """
        :return: Cardinality of intersection of `bgrms1` and `bgrms2` multisets
        """
        assert isinstance(bgrms1, BigramMultiset)
        assert isinstance(bgrms2, BigramMultiset)
        return libpysdcxx.wbigram_multiset_intersect_size(bgrms1._impl, bgrms2._impl)

    @staticmethod
    def sorensen_dice_coef(bgrms1: BigramMultiset, bgrms2: BigramMultiset) -> float:
        """
        :return: Sørensen–Dice coefficient of `bgrms1` and `bgrms2` multisets
        """
        assert isinstance(bgrms1, BigramMultiset)
        assert isinstance(bgrms2, BigramMultiset)
        return libpysdcxx.wbigram_multiset_sorensen_dice_coef(bgrms1._impl, bgrms2._impl)

    def __str__(self):
        return serialise(
            self,
            BigramMultiset._str_fixed_len + 4 * len(self),
            libpysdcxx.wbigram_multiset_str,
        )

    def __del__(self):
        libpysdcxx.delete_wbigram_multiset(self._impl)



import ctypes
from ctypes.util import find_library
from os import environ
from os.path import dirname, realpath
from glob import glob


def _load_libpysdcxx() -> ctypes.CDLL:
    """
    Load libpysdcxx library and set binding code up
    :return: Library functions handle
    """
    shared_obj = find_library("pysdcxx")
    if not shared_obj:  # huh, won't be that easy... :-(
        lib_dir = dirname(realpath(__file__))
        pysdcxx_libs = glob(f"{lib_dir}/../libpysdcxx.*")
        shared_obj = pysdcxx_libs[0] if pysdcxx_libs else None

    if not shared_obj:  # let's try LD_LIBRARY_PATH, I'm beginning to panic... :-((
        for ld_lib_dir in (environ.get("LD_LIBRARY_PATH") or "").split(':'):
            if not ld_lib_dir:
                continue

            pysdcxx_libs = glob(f"{ld_lib_dir}/libpysdcxx.*")
            if pysdcxx_libs:
                shared_obj = pysdcxx_libs[0]
                break

    if not shared_obj:  # bugger! x-(
        raise FileNotFoundError("Failed to find extension library")

    return ctypes.cdll.LoadLibrary(shared_obj)


def _bind_bigrams(libpysdcxx: ctypes.CDLL):
    # Constructors
    libpysdcxx.new_wbigrams.restype = ctypes.c_void_p

    libpysdcxx.new_wbigrams_str.argtypes = (ctypes.c_wchar_p, )
    libpysdcxx.new_wbigrams_str.restype = ctypes.c_void_p

    libpysdcxx.new_wbigrams_copy.argtypes = (ctypes.c_void_p, )
    libpysdcxx.new_wbigrams_copy.restype = ctypes.c_void_p

    # Destructor
    libpysdcxx.delete_wbigrams.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_wbigrams.restype = None  # void

    # Size
    libpysdcxx.wbigrams_size.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigrams_size.restype = ctypes.c_size_t

    # Iterators
    libpysdcxx.wbigrams_cbegin.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigrams_cbegin.restype = ctypes.c_void_p

    libpysdcxx.wbigrams_cend.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigrams_cend.restype = ctypes.c_void_p

    libpysdcxx.wbigrams_citer_ne.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigrams_citer_ne.restype = ctypes.c_int

    libpysdcxx.wbigrams_citer_deref.argtypes = (
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.POINTER(ctypes.c_size_t),
    )
    libpysdcxx.wbigrams_citer_deref.restype = None  # void

    libpysdcxx.wbigrams_citer_inc.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigrams_citer_inc.restype = None  # void

    libpysdcxx.delete_wbigrams_citer.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_wbigrams_citer.restype = None  # void

    # Union (add operators)
    libpysdcxx.wbigrams_iadd.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigrams_iadd.restype = ctypes.c_void_p

    libpysdcxx.wbigrams_add.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigrams_add.restype = ctypes.c_void_p

    # SDC
    libpysdcxx.wbigrams_intersect_size.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigrams_intersect_size.restype = ctypes.c_size_t

    libpysdcxx.wbigrams_sorensen_dice_coef.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigrams_sorensen_dice_coef.restype = ctypes.c_double

    # Serialisation
    libpysdcxx.wbigrams_str.argtypes = (
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.c_size_t,
    )
    libpysdcxx.wbigrams_str.restype = ctypes.c_size_t


def _bind_bigram_multiset(libpysdcxx: ctypes.CDLL):
    # Constructors
    libpysdcxx.new_wbigram_multiset.restype = ctypes.c_void_p

    libpysdcxx.new_wbigram_multiset_str.argtypes = (ctypes.c_wchar_p, )
    libpysdcxx.new_wbigram_multiset_str.restype = ctypes.c_void_p

    libpysdcxx.new_wbigram_multiset_copy.argtypes = (ctypes.c_void_p, )
    libpysdcxx.new_wbigram_multiset_copy.restype = ctypes.c_void_p

    # Destructor
    libpysdcxx.delete_wbigram_multiset.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_wbigram_multiset.restype = None  # void

    # Size
    libpysdcxx.wbigram_multiset_size.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigram_multiset_size.restype = ctypes.c_size_t

    # Iterators
    libpysdcxx.wbigram_multiset_cbegin.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigram_multiset_cbegin.restype = ctypes.c_void_p

    libpysdcxx.wbigram_multiset_cend.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigram_multiset_cend.restype = ctypes.c_void_p

    libpysdcxx.wbigram_multiset_citer_ne.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigram_multiset_citer_ne.restype = ctypes.c_int

    libpysdcxx.wbigram_multiset_citer_deref.argtypes = (
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.POINTER(ctypes.c_wchar),
    )
    libpysdcxx.wbigram_multiset_citer_deref.restype = None  # void

    libpysdcxx.wbigram_multiset_citer_inc.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wbigram_multiset_citer_inc.restype = None  # void

    libpysdcxx.delete_wbigram_multiset_citer.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_wbigram_multiset_citer.restype = None  # void

    # Union (add operators)
    libpysdcxx.wbigram_multiset_iadd.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigram_multiset_iadd.restype = ctypes.c_void_p

    libpysdcxx.wbigram_multiset_add.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wbigram_multiset_add.restype = ctypes.c_void_p

    # SDC
    libpysdcxx.wbigram_multiset_intersect_size.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
    )
    libpysdcxx.wbigram_multiset_intersect_size.restype = ctypes.c_size_t

    libpysdcxx.wbigram_multiset_sorensen_dice_coef.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
    )
    libpysdcxx.wbigram_multiset_sorensen_dice_coef.restype = ctypes.c_double

    # Serialisation
    libpysdcxx.wbigram_multiset_str.argtypes = (
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.c_size_t,
    )
    libpysdcxx.wbigram_multiset_str.restype = ctypes.c_size_t


def _bind_unordered_bigram_multiset(libpysdcxx: ctypes.CDLL):
    # Constructors
    libpysdcxx.new_unordered_wbigram_multiset.restype = ctypes.c_void_p

    libpysdcxx.new_unordered_wbigram_multiset_str.argtypes = (ctypes.c_wchar_p, )
    libpysdcxx.new_unordered_wbigram_multiset_str.restype = ctypes.c_void_p

    libpysdcxx.new_unordered_wbigram_multiset_copy.argtypes = (ctypes.c_void_p, )
    libpysdcxx.new_unordered_wbigram_multiset_copy.restype = ctypes.c_void_p

    # Destructor
    libpysdcxx.delete_unordered_wbigram_multiset.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_unordered_wbigram_multiset.restype = None  # void

    # Size
    libpysdcxx.unordered_wbigram_multiset_size.argtypes = (ctypes.c_void_p, )
    libpysdcxx.unordered_wbigram_multiset_size.restype = ctypes.c_size_t

    # Iterators
    libpysdcxx.unordered_wbigram_multiset_cbegin.argtypes = (ctypes.c_void_p, )
    libpysdcxx.unordered_wbigram_multiset_cbegin.restype = ctypes.c_void_p

    libpysdcxx.unordered_wbigram_multiset_cend.argtypes = (ctypes.c_void_p, )
    libpysdcxx.unordered_wbigram_multiset_cend.restype = ctypes.c_void_p

    libpysdcxx.unordered_wbigram_multiset_citer_ne.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
    )
    libpysdcxx.unordered_wbigram_multiset_citer_ne.restype = ctypes.c_int

    libpysdcxx.unordered_wbigram_multiset_citer_deref.argtypes = (
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.POINTER(ctypes.c_wchar),
    )
    libpysdcxx.unordered_wbigram_multiset_citer_deref.restype = None  # void

    libpysdcxx.unordered_wbigram_multiset_citer_inc.argtypes = (ctypes.c_void_p, )
    libpysdcxx.unordered_wbigram_multiset_citer_inc.restype = None  # void

    libpysdcxx.delete_unordered_wbigram_multiset_citer.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_unordered_wbigram_multiset_citer.restype = None  # void

    # Union (add operators)
    libpysdcxx.unordered_wbigram_multiset_iadd.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
    )
    libpysdcxx.unordered_wbigram_multiset_iadd.restype = ctypes.c_void_p

    libpysdcxx.unordered_wbigram_multiset_add.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
    )
    libpysdcxx.unordered_wbigram_multiset_add.restype = ctypes.c_void_p

    # SDC
    libpysdcxx.unordered_wbigram_multiset_intersect_size.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
    )
    libpysdcxx.unordered_wbigram_multiset_intersect_size.restype = ctypes.c_size_t

    libpysdcxx.unordered_wbigram_multiset_sorensen_dice_coef.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
    )
    libpysdcxx.unordered_wbigram_multiset_sorensen_dice_coef.restype = ctypes.c_double

    # Serialisation
    libpysdcxx.unordered_wbigram_multiset_str.argtypes = (
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.c_size_t,
    )
    libpysdcxx.unordered_wbigram_multiset_str.restype = ctypes.c_size_t


def _bind_sequence_matcher(libpysdcxx: ctypes.CDLL):
    # Constructor
    libpysdcxx.new_wsequence_matcher.restype = ctypes.c_void_p

    # Destructor
    libpysdcxx.delete_wsequence_matcher.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_wsequence_matcher.restype = None  # void

    # Reserve space for sequence
    libpysdcxx.wsequence_matcher_reserve.argtypes = (ctypes.c_void_p, ctypes.c_size_t)
    libpysdcxx.wsequence_matcher_reserve.restype = None  # void

    # Size
    libpysdcxx.wsequence_matcher_size.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_size.restype = ctypes.c_size_t

    # Push bigrams back
    libpysdcxx.wsequence_matcher_push_back.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
        ctypes.c_int,
    )
    libpysdcxx.wsequence_matcher_push_back.restype = None  # void

    # Push in-place created bigrams back
    libpysdcxx.wsequence_matcher_emplace_back.argtypes = (
        ctypes.c_void_p,
        ctypes.c_wchar_p,
        ctypes.c_int,
    )
    libpysdcxx.wsequence_matcher_emplace_back.restype = None  # void

    # Match iterators
    libpysdcxx.wsequence_matcher_begin.argtypes = (
        ctypes.c_void_p,
        ctypes.c_void_p,
        ctypes.c_double,
    )
    libpysdcxx.wsequence_matcher_begin.restype = ctypes.c_void_p

    libpysdcxx.wsequence_matcher_end.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_end.restype = ctypes.c_void_p

    libpysdcxx.wsequence_matcher_iter_deref.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_iter_deref.restype = ctypes.c_void_p

    libpysdcxx.wsequence_matcher_iter_size.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_iter_size.restype = ctypes.c_size_t

    libpysdcxx.wsequence_matcher_iter_begin.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_iter_begin.restype = ctypes.c_size_t

    libpysdcxx.wsequence_matcher_iter_end.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_iter_end.restype = ctypes.c_size_t

    libpysdcxx.wsequence_matcher_iter_sdc.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_iter_sdc.restype = ctypes.c_double

    libpysdcxx.wsequence_matcher_iter_inc.argtypes = (ctypes.c_void_p, )
    libpysdcxx.wsequence_matcher_iter_inc.restype = None  # void

    libpysdcxx.wsequence_matcher_iter_ne.argtypes = (ctypes.c_void_p, ctypes.c_void_p)
    libpysdcxx.wsequence_matcher_iter_ne.restype = ctypes.c_int

    libpysdcxx.delete_wsequence_matcher_iter.argtypes = (ctypes.c_void_p, )
    libpysdcxx.delete_wsequence_matcher_iter.restype = None  # void

    # Serialise match iterator
    libpysdcxx.wsequence_matcher_iter_str.argtypes = (
        ctypes.c_void_p,
        ctypes.c_wchar_p,
        ctypes.c_size_t,
    )
    libpysdcxx.wsequence_matcher_iter_str.restype = ctypes.c_size_t


libpysdcxx = _load_libpysdcxx()
_bind_bigrams(libpysdcxx)
_bind_bigram_multiset(libpysdcxx)
_bind_unordered_bigram_multiset(libpysdcxx)
_bind_sequence_matcher(libpysdcxx)



from ctypes import create_unicode_buffer


def serialise(obj: object, max_len: int, impl):
    """
    Object serialisation helper function

    The function calls native code serialisation implementation, providing it with
    string buffer for the result.
    The produced string is expected to fit in `max_len` characters.
    The native code honours that maximum and returns the number of characters
    written.
    In case of incomplete serialisation, the function adds dots at the end of
    produced string to indicate that situation.
    (The serialisation should never be used for anything serious; just in logs etc.)

    :param obj: Serialised object
    :param max_len: Maximal expected serialisation string length
    :param impl: `ctypes` implementation function pointer
    :return: Serialisation string
    """
    buf = create_unicode_buffer(max_len + 4)  # include space for "...\0"
    written = impl(obj._impl, buf, max_len)
    if not written < max_len:
        buf[max_len + 0] = '.'
        buf[max_len + 1] = '.'
        buf[max_len + 2] = '.'
        buf[max_len + 3] = '\0'

    return f"{obj.__class__.__name__}.{buf.value}"



#!/usr/bin/env python3

from distutils.core import setup, Extension
from os.path import dirname, realpath
import re


version = dict()
with open(f"{dirname(realpath(__file__))}/version.txt") as version_fd:
    for line in version_fd:
        level, no = re.fullmatch(r'^\s*VERSION_([A-Z]+)\s+([0-9]+)\s*$', line).groups()
        version[level.lower()] = int(no)


setup(name="pysdcxx",
    version=f"{version['major']}.{version['minor']}.{version['patch']}",
    description="Sørensen–Dice coefficient on string bigram multi-sets (in C++)",
    author="Václav Krpec",
    author_email="vencik@razdva.cz",
    url="https://github.com/vencik/libsdcxx",
    license="BSD-3-Clause license",
    license_files=["LICENSE"],
    package_dir={"": "src"},
    packages=["pysdcxx"],
    ext_modules=[Extension(
        "libpysdcxx",
        sources=[
            "src/libpysdcxx/bigrams.cxx",
            "src/libpysdcxx/bigram_multiset.cxx",
            "src/libpysdcxx/unordered_bigram_multiset.cxx",
            "src/libpysdcxx/sequence_matcher.cxx",
        ],
        extra_compile_args=["-Isrc", "-std=c++17"],
    )],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Text Processing",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3",
    ],
)



from __future__ import annotations
from typing import Optional, ClassVar, Tuple, Union, Type
from dataclasses import dataclass
import ctypes

from .libpysdcxx import libpysdcxx
from .bigrams import Bigrams


class SequenceMatcher:
    """
    String (token) sequence matcher.

    Uses Sørensen–Dice coefficient calculated for bigram mnultisets of (sequences of)
    tokens to find "fuzzy" matches in given text.
    """

    class Error(Exception):
        """
        Sequence matcher error
        """

    TokenOrBigrams = Union[str, Bigrams]                        # token or token Bigrams
    Token = Union[TokenOrBigrams, Tuple[TokenOrBigrams, bool]]  # token (with strip flag)

    @dataclass
    class Match:
        """
        Token sub-sequence match
        :param begin: Index of the 1st token of the matching sub-sequence
        :param end: Index just past the last token of the matching sub-sequence
        :param score: Sørensen–Dice similarity of the sub-sequence
        :param bigrams: Matching sub-sequence bigrams (if required)
        """
        begin: int
        end: int
        score: float
        bigrams: Optional[Bigrams]

    def __init__(
        self,
        tokens: Optional[Iterable[Token]] = None,
        reserve: int = 0,
    ):
        """
        The matcher manages upper triangular matrix of token bigrams.
        If `tokens` parameter is provided and its size can be found then reservation
        of space for the token bigrams is done automatically (e.g. if it's a `list`).
        In other cases, the `reserve` parameter may be used for that purpose.
        Note that space reservation is not required; yet may notably speed up
        construction of the matcher.

        The `tokens` parameter is expected to iterate the sequence tokens.
        The following values are acceptable:
        * `str` token
        * `Bigrams` object
        * `tuple[str,bool]` of `str` token together with "strip" flag (see `append`)
        * `tuple[Bigrams,bool]` the same for `Bigrams` object

        :param tokens: Token sequence
        :param reserve: Reserve space for `reserve` tokens of text (for 1-by-1 additions)
        """
        self._impl = libpysdcxx.new_wsequence_matcher()

        if tokens and hasattr(tokens, "__len__"):
            reserve = len(tokens)
        self.reserve(reserve)

        if tokens:
            for token in tokens:
                strip = False
                if isinstance(token, tuple):
                    token, strip = token

                self.append(token, strip)

    def reserve(self, size: int):
        """
        Reserve space for token bigrams
        :param size: Anticipated number of tokens
        """
        libpysdcxx.wsequence_matcher_reserve(self._impl, size)

    def append(self, token: TokenOrBigrams, strip: bool = False):
        """
        Append another `token` to the sequence

        If the `strip` parameter is set to true, the matcher shall not produce matches
        which begin or end with that token.

        :param token: Appended token (may be a string or `Bigrams` object)
        :param strip: Whether matches may begin/ end by the token or not
        """
        strip_int = 1 if strip else 0

        if isinstance(token, Bigrams):
            libpysdcxx.wsequence_matcher_push_back(self._impl, token._impl, strip_int)
        elif isinstance(token, str):
            libpysdcxx.wsequence_matcher_emplace_back(self._impl, token, strip_int)
        else:
            raise SequenceMatcher.Error(f"Unsupported token: {token}")

    def __len__(self):
        """
        :return: Number of token bigrams (i.e. full sequence length)
        """
        return libpysdcxx.wsequence_matcher_size(self._impl)

    def __deepcopy__(self, memo):
        """
        Make a copy on the native level
        :param memo: IDs of already copied objects (unused, we're non-recursive)
        """
        raise SequenceMatcher.Error(
            f"Copying {self.__class__.__name__} objects is not supported")

    def __copy__(self):
        """
        We don't do shallow copies
        """
        return self.__deepcopy__(None)

    def match(
        self,
        tokens: Union[TokenOrBigrams, Iterable[TokenOrBigrams]],
        threshold: float,
        include_bigrams: bool = False,
    ) -> Iterator[SequenceMatcher.Match]:
        """
        Match `tokens` to the matcher-managed token sequence

        The function can, again, accept various types for the `tokens` parameter:
        * A `str` token
        * A `Bigrams` object
        * An iterable of `str` tokens
        * An iterable of `Bigrams` objects

        Produced match may take 2 possible forms:
        * `tuple[int,int,float]`: a tuple containing match begin, end and score
        * `tuple[int,int,float,Bigrams]` same, enriched with matching sub-seq. bigrams

        Match begin is the index of matching sub-sequence 1st token.
        Match end is index just past matching sub-sequence last token (following usual
        Python begin-end iteration specification convention).
        Match score is the calculated Sørensen–Dice similarity of the matching
        sub-sequence and the specified `token`.
        If `include_bigrams` is `True`, the tuple shall also contain the matching
        sub-sequence bigrams.

        :param tokens: Matched tokens specification
        :param threshold: Matching score (Sørensen–Dice coefficient) threshold
        :param include_bigrams: Include matching sub-sequence bigrams in match tuple
        :return: Generator of matches
        """
        if isinstance(tokens, Bigrams):  # single bigram multiset
            bgrms = tokens

        elif isinstance(tokens, str):  # single token
            bgrms = Bigrams(tokens)

        elif hasattr(tokens, "__iter__"):  # create Bigrams union
            bgrms = Bigrams()
            for token in tokens:
                if isinstance(token, Bigrams):
                    pass
                elif isinstance(token, str):
                    token = Bigrams(token)
                else:
                    raise SequenceMatcher.Error(f"Unsupported token: {token}")

                bgrms += token

        else:
            raise SequenceMatcher.Error(f"Unsupported token: {token}")

        itr = libpysdcxx.wsequence_matcher_begin(self._impl, bgrms._impl, threshold)
        end = libpysdcxx.wsequence_matcher_end(self._impl)
        try:
            while libpysdcxx.wsequence_matcher_iter_ne(itr, end):
                yield SequenceMatcher.Match(
                    begin=libpysdcxx.wsequence_matcher_iter_begin(itr),
                    end=libpysdcxx.wsequence_matcher_iter_end(itr),
                    score=libpysdcxx.wsequence_matcher_iter_sdc(itr),
                    bigrams=Bigrams(_impl=libpysdcxx.wsequence_matcher_iter_deref(itr)) \
                        if include_bigrams else None,
                )

                libpysdcxx.wsequence_matcher_iter_inc(itr)

        finally:
            libpysdcxx.delete_wsequence_matcher_iter(end)
            libpysdcxx.delete_wsequence_matcher_iter(itr)

    def __del__(self):
        libpysdcxx.delete_wsequence_matcher(self._impl)
